class AugerConfig(object):
    def __init__(self, ctx):
        super(AugerConfig, self).__init__()
        self.ctx = ctx

    def set_data_set(self, name, source=None, validation=False, user_name=None):
        #TODO: add more providers later
        if validation:
            self.ctx.config.set('experiment/validation_dataset', name)
            if self.ctx.use_auger_cloud() and 'azure' in self.ctx.get_providers():
                self.ctx.config.set('experiment/validation_dataset', name, "azure")
        else:
            #print("set_data_set: %s"%self.ctx.use_auger_cloud())
            self.ctx.config.set('dataset', name)
            if user_name:
                self.ctx.config.set('dataset_name', user_name)
                self.ctx.config.set(f'experiments/{user_name}/dataset', self.ctx.config.get('dataset'))

            if self.ctx.use_auger_cloud() and 'azure' in self.ctx.get_providers():
                self.ctx.config.set('dataset', name, "azure")

        self.ctx.config.write_all()
        return self
            
    def set_experiment(self, experiment_name, experiment_session_id):
        self.ctx.config.set('experiment/name', experiment_name)
        self.ctx.config.set('experiment/experiment_session_id', experiment_session_id)

        if self.ctx.config.get('dataset_name'):
            dataset_name = self.ctx.config.get('dataset_name')
            self.ctx.config.set(f'experiments/{dataset_name}/experiment_name', experiment_name)
            self.ctx.config.set(f'experiments/{dataset_name}/experiment_session_id', experiment_session_id)
            self.ctx.config.set(f'experiments/{dataset_name}/dataset', self.ctx.config.get('dataset'))

        self.ctx.config.write()

    def _get_experiment_by_dataset(self):
        dataset_name = self.ctx.config.get('dataset_name')
        experiments = self.ctx.config.get('experiments', {})
        
        return experiments.get(dataset_name, {})

    def get_experiment(self):                
        return self._get_experiment_by_dataset().get('experiment_name',
            self.ctx.config.get('experiment/name'))

    def get_experiment_session(self):                
        return self._get_experiment_by_dataset().get('experiment_session_id',
            self.ctx.config.get('experiment/experiment_session_id'))

    def get_dataset(self):
        return self._get_experiment_by_dataset().get('dataset',
            self.ctx.config.get('dataset'))
            
    def set_project(self, project_name):
        self.ctx.config.set('name', project_name)
        self.ctx.config.write()
        return self




import os
import pytest
import shutil
import logging
import json

from click.testing import CliRunner
from a2ml.api.utils.context import Context


@pytest.fixture
def ctx():
    # load config(s) from the test app
    return Context(debug=True)

@pytest.fixture
def runner():
    return CliRunner()

@pytest.fixture(scope="function")
def isolated(runner):
    with runner.isolated_filesystem():
        yield runner

@pytest.fixture
def log(caplog):
    caplog.set_level(logging.INFO)
    return caplog

@pytest.fixture
def project(isolated):
    source = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'fixtures',
        'cli-integration-test')
    shutil.copytree(source, './cli-integration-test')
    os.chdir('cli-integration-test')

TEST_AUGER_CREDENTIALS = {
    'username': 'test_user',
    'organization': 'auger',
    'api_url': 'https://example.com',
    'token': 'fake_token',
}

@pytest.fixture
def auger_authenticated(monkeypatch, isolated):
    monkeypatch.setenv("AUGER_CREDENTIALS", json.dumps(TEST_AUGER_CREDENTIALS))
    #monkeypatch.setenv("AUGER_CREDENTIALS_PATH", os.getcwd())



import os
import json
import shutil
from a2ml.api.utils.base_credentials import BaseCredentials
from .exceptions import AzureException
from a2ml.api.utils import fsclient


class Credentials(BaseCredentials):
    """Manage credentials on user computer."""
    def __init__(self, ctx):
        super(Credentials, self).__init__(ctx, "azure")
        self.subscription_id = None
        self.directory_tenant_id = None
        self.application_client_id = None
        self.client_secret = None

    def load(self):
        content = {}
        if hasattr(self.ctx, 'credentials') and self.ctx.credentials:
            content = self.ctx.credentials
        elif 'AZURE_CREDENTIALS' in os.environ:
            content = os.environ.get('AZURE_CREDENTIALS', None)
            content = json.loads(content) if content else {}
        else:
            #look for service principal credentials
            if self._credentials_file_exist():
                with open(self.creds_file, 'r') as file:
                    content = json.loads(file.read())
            else:
                content = self._load_azure_cred_file()

                if not content.get('subscription_id') and not self.ctx.config.get('use_server'):
                    from azureml.core.authentication import InteractiveLoginAuthentication
                    #fallback to force browser login for token
                    interactive_auth = InteractiveLoginAuthentication(force=True)
                    interactive_auth.get_authentication_header()

                    content = self._load_azure_cred_file()
                                
        self.subscription_id = content.get('subscription_id')
        self.directory_tenant_id = content.get('directory_tenant_id')
        self.application_client_id = content.get('application_client_id')
        self.client_secret = content.get('client_secret')

        return self

    def _load_azure_cred_file(self):
        content = {}

        azure_creds_file = os.path.abspath('%s/.azureml/auth/azureProfile.json' % os.environ.get('HOME', ''))
        if os.path.exists(azure_creds_file):
            from a2ml.api.utils import fsclient
            try:
                with fsclient.open_file(azure_creds_file, "r", encoding='utf-8-sig', num_tries=0) as file:
                    res = json.load(file)
                    content = {
                        'subscription_id': res['subscriptions'][0]['id']
                    }
            except Exception as e:
                if self.ctx.debug:
                    import traceback
                    traceback.print_exc()
        
        return content

    def login(self, username, password, organization, url=None):
        self.load()
        try:
            self.save()
            self.ctx.log(
                'You are now logged in to Azure')
        except Exception as exc:
            exc_text = str(exc)
            self.ctx.log(exc_text)

    def logout(self):
        self.load()
        azure_root_dir = os.path.abspath('%s/.azureml' % os.environ.get('HOME', ''))
        fsclient.remove_folder(azure_root_dir)
        self.subscription_id = None
        self.directory_tenant_id = None
        self.application_client_id = None
        self.client_secret = None
        self.save()
        self.ctx.log('You are logged out of Azure.')

    def whoami(self):
        self.load()
        if self.subscription_id is None:
            self.ctx.log('Please login to Azure...')
        else:
            self.ctx.log(
                'subscription_id: %s directory_tenant_id:%s application_client_id:%s client_secret:%s' % (
                    self.subscription_id,
                    self.directory_tenant_id,
                    self.application_client_id,
                    self.client_secret))

    def serialize(self):
        return {
            'subscription_id' : self.subscription_id,
            'directory_tenant_id' : self.directory_tenant_id,
            'application_client_id': self.application_client_id,
            'client_secret': self.client_secret
        }

    def get_serviceprincipal_auth(self):
        from azureml.core.authentication import ServicePrincipalAuthentication

        svc_pr = None
        if self.client_secret:
            svc_pr = ServicePrincipalAuthentication(
                tenant_id=self.directory_tenant_id,
                service_principal_id=self.application_client_id,
                service_principal_password=self.client_secret
            )
        return svc_pr

    def save(self):
        self._ensure_credentials_file()

        content = self.serialize()
        with open(self.creds_file, 'w') as file:
            file.write(json.dumps(content))

    def verify(self):
        if self.ctx.config.get('use_server'):
            if not self.subscription_id or \
               not self.directory_tenant_id or \
               not self.application_client_id or \
               not self.client_secret:
               raise AzureException('Please provide azure.json file with Azure AD application and service principal.')

        elif self.subscription_id is None:
            raise AzureException('Please provide your credentials to Azure...')

        return True



import os

from a2ml.api.a2ml import A2ML
from a2ml.api.utils.context import Context
from a2ml.api.utils.dataframe import DataFrame


def test_text_preprocess():

    a2mlObj = A2ML(Context())

    dsEmp =  DataFrame.create_dataframe(os.path.abspath('./tests/fixtures/local/employees.csv'))

    params = {
        'text_cols': ['GF_GOAL', 'GF_EMP_COMMENT'],
        'text_metrics': ['mean_length', 'unique_count', 'separation_score'],
        'output_prefix': 'empl_',

        #'vectorize': 'en_use_lg',
        'vectorize': 'hashing',
        'dim_reduction': {'alg_name': 'PCA', 'args': {'n_components': 2}}
    }

    res = a2mlObj.preprocess_data(dsEmp.df, [{'text': params}], locally=True)

    assert res['result']
    assert len(res['data'])
    assert res['data'].columns.tolist()==['EMPLID', 'GF_COMMENTS1', 'GF_MGR_COMMENT', 'GF_JOB_TITLE','extractions', 'SIGNAL_LEN', 'JOB_TYPE', 'empl_GF_GOAL_0','empl_GF_GOAL_1', 'empl_GF_EMP_COMMENT_0', 'empl_GF_EMP_COMMENT_1']


import math
import random
from inspect import getfullargspec


class BaseInterpreter(object):
    MAX_ARGS_COUNT = 255

    def snake_case(self, str):
        res = ''

        for i, c in enumerate(str):
            if c.isupper() and i > 0:
                res += "_"

            res += c.lower()

        return res

    def node_method_name(self, node):
        return 'evaluate_' + self.snake_case(type(node).__name__)

    def evaluate(self, node, rows=None):
        evaluateor = getattr(self, self.node_method_name(node), self.generic_evaluate)

        if rows is None:
            return evaluateor(node)
        else:
            return evaluateor(node, rows)

    def generic_evaluate(self, node):
        raise Exception('No {} method'.format(self.node_method_name(node)))

    # Builtin functions

    @staticmethod
    def logic_if(interpreter, predicate, true_value, false_value):
        if interpreter.evaluate(predicate):
            return interpreter.evaluate(true_value)
        else:
            return interpreter.evaluate(false_value)

    @staticmethod
    def agg_max(_, args):
        return max(args)

    @staticmethod
    def agg_min(_, args):
        return min(args)

    @staticmethod
    def log(_, x, base=math.e):
        return math.log(x, base)

    @staticmethod
    def min(_, arg1, arg2, *args):
        return min(arg1, arg2, *args)

    @staticmethod
    def max(_, arg1, arg2, *args):
        return max(arg1, arg2, *args)

    @staticmethod
    def randint(_, a, b):
        return random.randint(a, b)

    @staticmethod
    def random(_):
        return random.random()

    @staticmethod
    def func_values():
        return {
            # Logic
            "if": BaseInterpreter.logic_if,
            "@if": BaseInterpreter.logic_if,

            # Aggregation
            "agg_max": BaseInterpreter.agg_max,
            "agg_min": BaseInterpreter.agg_min,

            # Math
            "abs": abs,
            "ceil": math.ceil,
            "cos": math.cos,
            "exp": math.exp,
            "floor": math.floor,
            "log10": math.log10,
            "log2": math.log2,
            "round": round,
            "sin": math.sin,
            "sqrt": math.sqrt,
            "tan": math.tan,

            # Str
            "len": len,

            # For these builtin functions getfullargspec doesn't work so redefine them as a staticmethods
            "log": BaseInterpreter.log,
            "max": BaseInterpreter.max,
            "min": BaseInterpreter.min,
            "randint": BaseInterpreter.randint,
            "random": BaseInterpreter.random,
        }

    @staticmethod
    def is_static_func(func):
        return func.__qualname__.startswith(f"{BaseInterpreter.__name__}.")

    @staticmethod
    def known_funcs():
        res = {}
        funcs = BaseInterpreter.func_values()

        for func_name, func in funcs.items():
            # All static functions defined in BaseInterpreter accept interpreter instance as 1 parameter
            if BaseInterpreter.is_static_func(func):
                self_arg_count = 1
            else:
                self_arg_count = 0

            spec = getfullargspec(func)

            # -1 for interpreter it self
            # -len(spec.defaults) for default arg values
            min_count = len(spec.args) - self_arg_count - len(spec.defaults or [])
            max_count = len(spec.args) - self_arg_count

            if spec.varargs:
                max_count = BaseInterpreter.MAX_ARGS_COUNT

            res[func_name] = range(min_count, max_count + 1)

        return res



import os
import json
import time

from .exceptions import AzureException
from a2ml.api.utils.dataframe import DataFrame
from a2ml.api.utils import fsclient, retry_helper
from a2ml.api.utils.decorators import error_handler, authenticated
from a2ml.api.model_review.model_helper import ModelHelper
from a2ml.api.model_review.model_review import ModelReview
from .credentials import Credentials


class AzureModel(object):

    def __init__(self, ctx):
        super(AzureModel, self).__init__()
        self.ctx = ctx
        self.credentials = Credentials(self.ctx).load()

    @error_handler
    @authenticated
    def deploy(self, model_id, locally, review, name=None, algorithm=None, score=None, data_path=None):
        if locally:
            is_loaded, model_path = self.verify_local_model(model_id)
            if is_loaded:
                self.ctx.log('Model already deployed to %s' % model_path)
                return {'model_id': model_id}

        from azureml.train.automl.run import AutoMLRun

        ws, experiment = self._get_experiment()
        model_run = AutoMLRun(experiment = experiment, run_id = model_id)

        result = self._deploy_locally(model_id, model_run, ws, experiment) if locally else \
            self._deploy_remotly(model_id, model_run, ws, experiment)
        result = True

        model_features, target_categories = self._get_remote_model_features(model_run)
        feature_importance = self._get_feature_importance(model_run)

        options = {
            'uid': model_id,
            'targetFeature': self.ctx.config.get('target'),
            'support_review_model': review,
            'provider': self.ctx.config.name,
            'scoreNames': [self.ctx.config.get('experiment/metric')],
            'scoring': self.ctx.config.get('experiment/metric'),
            "score_name": self.ctx.config.get('experiment/metric'),
            "originalFeatureColumns": model_features,
            "model_type": self.ctx.config.get("model_type")
        }
        options.update(self._get_hub_info())
        fsclient.write_json_file(os.path.join(self.ctx.config.get_model_path(model_id), "options.json"),
            options)
        target_categoricals = {}
        if target_categories:
            target_categoricals = {self.ctx.config.get('target'): {"categories": target_categories}}
        fsclient.write_json_file(os.path.join(self.ctx.config.get_model_path(model_id), "target_categoricals.json"),
            target_categoricals )

        metric_path = ModelHelper.get_metric_path( options, model_id)
        if metric_path:
            fsclient.write_json_file(os.path.join(metric_path, "metric_names_feature_importance.json"),
                {'feature_importance_data': {
                    'features': list(feature_importance.keys()),
                    'scores': list(feature_importance.values())
                }})

        return result

    def _get_hub_info(self):
        return {'hub_info':{
                'project_path': self.ctx.config.get_path(),
                'experiment_id': self.ctx.config.get('experiment/hub_experiment_id', None),
                'experiment_session_id':self.ctx.config.get('experiment/hub_session_id', None),
            }};

    def _deploy_remotly(self, model_id, model_run, ws, experiment):
        from azureml.core.model import Model
        from azureml.core.model import InferenceConfig
        from azureml.train.automl.run import AutoMLRun

        # ws, experiment = self._get_experiment()
        iteration, run_id = self._get_iteration(model_id)

        experiment_run = AutoMLRun(experiment = experiment, run_id = run_id)
        model_name = model_run.properties['model_name']
        self.ctx.log('Registering model: %s' % model_id)

        description = '%s-%s' % (model_name, iteration)
        model = experiment_run.register_model(
            model_name = model_name, iteration=iteration,
            description = description, tags = None)

        script_file_name = '.azureml/score_script.py'
        model_run.download_file(
            'outputs/scoring_file_v_1_0_0.py', script_file_name)

        self._edit_score_script(script_file_name)

        service_name, service_config, service_target = \
            self._create_deploy_service(model_name, ws)

        inference_config = InferenceConfig(
            environment = model_run.get_environment(),
            entry_script = script_file_name)

        deploy_service = Model.deploy(
            ws, service_name, [model], inference_config, deployment_config=service_config,
            deployment_target=service_target, overwrite=True)
        deploy_service.wait_for_deployment(show_output=True)
        self.ctx.log('%s state %s' % (service_name, str(deploy_service.state)))

        return {'model_id': model_id, 'service_name': service_name}

    def _create_deploy_service(self, model_name, ws):
        from azureml.core.webservice import AciWebservice, AksWebservice, Webservice
        from azureml.exceptions import WebserviceException
        from azureml.core.compute import AksCompute, ComputeTarget
        from .project import AzureProject

        service_type = self.ctx.config.get('deploy_cluster/type', 'aci').lower()
        cpu_cores = float(self.ctx.config.get('deploy_cluster/cpu_cores', 1))
        memory_gb = float(self.ctx.config.get('deploy_cluster/memory_gb', 2))
        service_name = self._deploy_service_name(model_name)
        service_target = None

        if service_type == "aci":
            self.ctx.log('Deploying AciWebservice %s ...' % service_name)

            service_config = AciWebservice.deploy_configuration(
                cpu_cores = cpu_cores,
                memory_gb = memory_gb,
                tags = {'type': "inference-%s" % service_name},
                description = "inference-%s" % service_name)
        elif service_type == "aks":
            self.ctx.log('Deploying AksWebservice %s ...' % service_name)
            aks_compute_target = AzureProject._fix_cluster_name(
                self.ctx.config.get('deploy_cluster/compute_target', "a2ml_aks_deploy"))
            if not aks_compute_target in ws.compute_targets:
                if self.ctx.is_runs_on_server():
                    raise AzureException("Compute target %s does not exist."%aks_compute_target)

                self.ctx.log('Creating AksCompute %s ...' % aks_compute_target)
                prov_config = AksCompute.provisioning_configuration(
                    agent_count=int(self.ctx.config.get('deploy_cluster/agent_count', 2)),
                    vm_size=self.ctx.config.get('deploy_cluster/vm_size', "STANDARD_D2_V2"),
                    cluster_purpose=self.ctx.config.get('deploy_cluster/purpose', AksCompute.ClusterPurpose.DEV_TEST)
                )
                service_target = ComputeTarget.create(workspace = ws,
                                name = aks_compute_target,
                                provisioning_configuration = prov_config)

                # Wait for the create process to complete
                service_target.wait_for_completion(show_output = True)
            else:
                service_target = AksCompute(ws, aks_compute_target)

            service_config = AksWebservice.deploy_configuration(
                cpu_cores = cpu_cores,
                memory_gb = memory_gb)
        else:
            raise AzureException("deploy_cluster/type: %s it not supported. Supported type: aci, aks"%service_type)

        try:
            Webservice(ws, service_name).delete()
        except WebserviceException as exc:
            self.ctx.log("Delete existing service failed: %s"%exc.message)

        return service_name, service_config, service_target

    def _get_deploy_service(self, model_name, ws):
        from azureml.core.webservice import AciWebservice, AksWebservice

        service_name = self._deploy_service_name(model_name)
        service_type = self.ctx.config.get('deploy_cluster/type', 'aci').lower()

        if service_type == "aci":
            deploy_service = AciWebservice(ws, service_name)
        elif service_type == "aks":
            deploy_service = AksWebservice(ws, service_name)
        else:
            raise AzureException("deploy_cluster/type: %s it not supported. Supported type: aci, aks"%service_type)

        return deploy_service

    def _edit_score_script(self, script_file_name):
        text = fsclient.read_text_file(script_file_name)

        text = text.replace("@input_schema('data', PandasParameterType(input_sample))",
        """
def convert_simple_numpy_type(obj):
    if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
        np.int16, np.int32, np.int64, np.uint8,
        np.uint16, np.uint32, np.uint64)):
        return int(obj)
    elif isinstance(obj, (np.float_, np.float16, np.float32,
        np.float64)):
        return float(obj)

    return None

class NumpyJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        res = convert_simple_numpy_type(obj)
        if res is not None:
            return res

        return json.JSONEncoder.default(self, obj)

def json_dumps_np(data, allow_nan=True):
    return json.dumps(data, cls=NumpyJSONEncoder, allow_nan=allow_nan)

@input_schema('data', PandasParameterType(input_sample))
def get_df(data):
    return data

        """
        )

        text = text.replace("result = model.predict(data)",
        """
        df = get_df(data['data'])
        proba_classes = []
        if data['method'] == "predict_proba":
            result = model.predict_proba(df)
            proba_classes = list(model.classes_)
        else:
            result = model.predict(df)

        """
        )

        text = text.replace("return json.dumps({\"result\": result.tolist()})",
            "return json_dumps_np({\"result\": result.tolist(), \"proba_classes\": proba_classes})")
        fsclient.write_text_file(script_file_name, text)

    @error_handler
    @authenticated
    def predict(self, filename, model_id, threshold=None, locally=False, data=None, columns=None,
        predicted_at=None, output=None, json_result=False, count_in_result=False, prediction_id=None,
        no_features_in_result=None):
        ds = DataFrame.create_dataframe(filename, data, columns)
        model_path = self.ctx.config.get_model_path(model_id)
        options = fsclient.read_json_file(os.path.join(model_path, "options.json"))

        if threshold and options.get('model_type', 'classification') != 'classification':
            self.ctx.log("Threshold only applied to classification and will be ignored.")
            threshold = None

        results, results_proba, proba_classes, target_categories, model_features = \
            self._predict_locally(ds.df, model_id, threshold) if locally else self._predict_remotely(ds.df, model_id, threshold)

        if target_categories and len(target_categories) == 2:
            for idx, item in enumerate(target_categories):
                if item == "False":
                    target_categories[idx] = False
                if item == "True":
                    target_categories[idx] = True

        target_feature = options.get('targetFeature', self.ctx.config.get('target', None))
        ModelHelper.process_prediction(ds,
            results, results_proba, proba_classes,
            threshold,
            options.get('minority_target_class', self.ctx.config.get('minority_target_class')),
            target_feature,
            target_categories)

        gzip_predict_file = False
        if ds.count() > options.get('max_predict_records_to_gzip', 1000):
            gzip_predict_file = True

        if model_features:
            model_features += [target_feature]

        predicted = ModelHelper.save_prediction(ds, prediction_id,
            json_result, count_in_result, predicted_at,
            model_path, model_id, output, gzip_predict_file=gzip_predict_file, model_features=model_features)

        if filename:
            self.ctx.log('Predictions stored in %s' % predicted)

        return {'predicted': predicted}

    @error_handler
    @authenticated
    def actuals(self, model_id, filename=None, data=None, columns=None, actuals_at=None, actual_date_column=None, locally=False):
        if locally:
            model_path = self.ctx.config.get_model_path(model_id)

            if not fsclient.is_folder_exists(model_path):
                raise Exception('Model should be deployed first.')

            return ModelReview({'model_path': model_path}).add_actuals(
                self.ctx,
                actuals_path=filename,
                data=data,
                columns=columns,
                actual_date=actuals_at,
                actual_date_column=actual_date_column,
                provider='azure'
            )
        else:
            raise Exception("Not Implemented")

    @error_handler
    @authenticated
    def delete_actuals(self, model_id, with_predictions=False, begin_date=None, end_date=None, locally=False):
        if locally:
            model_path = self.ctx.config.get_model_path(model_id)

            if not fsclient.is_folder_exists(model_path):
                raise Exception('Model should be deployed first.')

            return ModelReview({'model_path': model_path}).delete_actuals(
                with_predictions=with_predictions, begin_date=begin_date, end_date=end_date)
        else:
            raise Exception("Not Implemented")

    @error_handler
    @authenticated
    def build_review_data(self, model_id, locally, output):
        if locally:
            model_path = self.ctx.config.get_model_path(model_id)

            if not fsclient.is_folder_exists(model_path):
                raise Exception('Model should be deployed first.')

            return ModelReview({'model_path': os.path.join(model_path, "model")}).build_review_data(
              data_path=self.ctx.config.get("source"), output=output)
        else:
            raise Exception("Not Implemented.")

    def _get_iteration(self, model_id):
        iteration = None
        run_id = model_id
        parts = model_id.split('_')
        if len(parts) > 2:
            run_id = parts[0]+"_"+parts[1]
            iteration = parts[2]
        return iteration, run_id

    def _deploy_service_name(self, model_name):
        # It must only consist of lowercase letters, numbers, or dashes, start
        # with a letter, end with a letter or number, and be between 3 and 32
        # characters long.
        #TODO - service_name + suffix must satisfy requiremets
        if model_name.endswith('-service'):
            return model_name

        return (model_name+'-service').lower()

    def _get_experiment(self):
        from azureml.core import Experiment
        from .project import AzureProject

        ws = AzureProject(self.ctx)._get_ws()
        experiment_name = self.ctx.config.get('experiment/name', None)
        if experiment_name is None:
            raise AzureException('Please specify Experiment name...')
        experiment = Experiment(ws, experiment_name)

        return ws, experiment

    def _get_remote_model_features(self, remote_run):
        from  a2ml.api.utils import fsclient, local_fsclient
        import pandas as pd

        model_features = None
        target_categories = None

        temp_dir = local_fsclient.LocalFSClient().get_temp_folder()
        try:
            file_name = 'scoring_file_v_1_0_0.py'
            remote_run.download_file('outputs/%s'%file_name, os.path.join(temp_dir, file_name))
            text = fsclient.read_text_file(os.path.join(temp_dir, file_name))
            to_find = "input_sample ="
            start = text.find(to_find)
            if start > 0:
                end = text.find("\n", start)
                if end > start:
                    code_to_run = text[start+len(to_find):end]

                    input_sample = eval(code_to_run)
                    model_features = input_sample.columns.tolist()
        except Exception as e:
            self.ctx.log('Cannot get columns from remote model.Use original columns from predicted data: %s'%e)

        if self.ctx.config.get("model_type") == "classification":
            try:
                file_name = 'confusion_matrix'
                remote_run.download_file('%s'%file_name, os.path.join(temp_dir, file_name))
                cm_data = fsclient.read_json_file(os.path.join(temp_dir, file_name))
                target_categories = cm_data.get('data', {}).get('class_labels')
            except Exception as e:
                self.ctx.log('Cannot get categorical target class labels from remote model.Use class codes: %s'%e)

        fsclient.remove_folder(temp_dir)
        return model_features, target_categories

    def _get_feature_importance(self, model_run):
        from azureml.interpret._internal.explanation_client import ExplanationClient

        try:
            client = ExplanationClient.from_run(model_run)
            engineered_explanations = client.download_model_explanation(raw=True)
            return engineered_explanations.get_feature_importance_dict()
        except Exception as e:
            self.ctx.log('Cannot get feature_importance from remote model: %s'%e)

        return {}

    def _call_service_run(self, deploy_service, input_data):
        return deploy_service.run(input_data = input_data)

    def _predict_remotely(self, predict_data, model_id, predict_proba):
        from azureml.train.automl.run import AutoMLRun
        from azureml.core.run import Run

        import numpy as np

        ws, experiment = self._get_experiment()

        model_features = None
        target_categories = None

        remote_run = AutoMLRun(experiment = experiment, run_id = model_id)
        model_features, target_categories = self._get_remote_model_features(remote_run)
        if model_id.startswith("AutoML_"):
            model_name = remote_run.properties['model_name']
        else:
            model_name = model_id

        if model_features:
            predict_data = predict_data[model_features]

        input_payload = predict_data.to_json(orient='split', index = False)
        input_payload = json.loads(input_payload)
        # If you have a classification model, you can get probabilities by changing this to 'predict_proba'.
        method = 'predict'
        if predict_proba:
            method = 'predict_proba'
        input_payload = {
            'data': {'data': input_payload['data'], 'method': method}
        }
        input_payload = json.dumps(input_payload)

        deploy_service = self._get_deploy_service(model_name, ws)

        try:
            response = retry_helper(lambda: self._call_service_run(deploy_service, input_payload),
                ['Connection aborted','Too many requests for service','WebserviceException'], ctx=self.ctx)
        except Exception as e:
            log_file = 'automl_errors.log'
            fsclient.write_text_file(log_file, deploy_service.get_logs(), mode="a")
            raise AzureException("Prediction service error. Please redeploy the model. Log saved to file '%s'. Details: %s"%(log_file, str(e)))

        response = json.loads(response)
        if "error" in response or not 'result' in response:
            raise AzureException('Prediction service return error: %s'%response.get('error'))

        results_proba = None
        proba_classes = None
        results = response['result']
        if predict_proba:
            results_proba = results
            proba_classes = response['proba_classes']
            results_proba = np.array(results_proba)

        return results, results_proba, proba_classes, target_categories, model_features

    def verify_local_model(self, model_id):
        model_path = os.path.join(self.ctx.config.get_model_path(model_id),
            'model.pkl.gz')
        return fsclient.is_file_exists(model_path), model_path

    def _deploy_locally(self, model_id, model_run, ws, experiment):
        from azureml.train.automl.run import AutoMLRun

        self.ctx.log('Downloading model %s' % model_id)

        iteration, run_id = self._get_iteration(model_id)
        remote_run = AutoMLRun(experiment = experiment, run_id = run_id)
        best_run, fitted_model = remote_run.get_output(iteration=iteration)

        is_loaded, model_path = self.verify_local_model(model_id)
        fsclient.save_object_to_file(fitted_model, model_path)

        # model_path = self.ctx.config.get_model_path(model_id)
        # with fsclient.save_local(os.path.join(model_path, 'model.pkl'), move_file=True) as local_path:
        #     model_run.download_file("outputs/model.pkl", local_path)

        self.ctx.log('Downloaded model to %s' % model_path)
        return {'model_id': model_id}

    def _predict_locally(self, predict_data, model_id, threshold):
        is_loaded, model_path = self.verify_local_model(model_id)
        if not is_loaded:
            raise Exception("Model should be deployed before predict.")

        fitted_model = fsclient.load_object_from_file(model_path, use_local_cache=True)
        model_features = None
        try:
            options = fsclient.read_json_file(os.path.join(self.ctx.config.get_model_path(model_id), "options.json"))

            model_features = options.get("originalFeatureColumns")
            predict_data = predict_data[model_features]
            predict_data.to_csv("test_options.csv", index=False, compression=None, encoding='utf-8')
        except Exception as e:
            self.ctx.log('Cannot get columns from model.Use original columns from predicted data: %s'%e)

        results_proba = None
        proba_classes = None
        results = None
        if threshold is not None:
            results_proba = fitted_model.predict_proba(predict_data)
            proba_classes = list(fitted_model.classes_)
        else:
            results = fitted_model.predict(predict_data)

        target_categoricals = fsclient.read_json_file(os.path.join(
                self.ctx.config.get_model_path(model_id), "target_categoricals.json"))
        target_categories = target_categoricals.get(self.ctx.config.get('target'), {}).get("categories")

        return results, results_proba, proba_classes, target_categories, model_features

    @error_handler
    @authenticated
    def undeploy(self, model_id, locally):
        if not self.ctx.config.get('undeploy/service_only'):
            model_path = self.ctx.config.get_model_path(model_id)
            self.ctx.log("Undeploy model. Remove model folder: %s"%model_path)
            fsclient.remove_folder(model_path)

        if not locally:
            from azureml.train.automl.run import AutoMLRun
            from azureml.core.webservice import Webservice
            from azureml.exceptions import WebserviceException

            ws, experiment = self._get_experiment()
            model_run = AutoMLRun(experiment = experiment, run_id = model_id)
            model_name = model_run.properties['model_name']

            service_name = self._deploy_service_name(model_name)
            try:
                Webservice(ws, service_name).delete()
            except WebserviceException as exc:
                #self.ctx.log(exc.message)
                raise AzureException(exc.message)

            self.ctx.log("Model endpoint has been removed.")

    @error_handler
    @authenticated
    def review_alert(self, model_id, parameters, name):
        raise AzureException("Not Implemented. Set use_auger_cloud: True in config.yml")

    @error_handler
    @authenticated
    def review(self, model_id):
        raise Exception("Not Implemented. Set use_auger_cloud: True in config.yml")



import os
import pytest
import json
from a2ml.cmdl.cmdl import cmdl

class TestAzureAuth():
    def test_login(self, log, runner, isolated, monkeypatch):
        azure_dict = {
          "subscription_id": "test_subscription_id",
          "directory_tenant_id": "test_directory_tenant_id",
          "application_client_id": "test_application_client_id",
          "client_secret": "test_client_secret"
        }
        AZURE_JSON = json.dumps(azure_dict)
        monkeypatch.setenv("AZURE_CREDENTIALS", AZURE_JSON)
        result = runner.invoke(
            cmdl,
            ['auth', 'login', '-p', 'azure'])
        assert result.exit_code == 0
        assert (log.records[-1].message ==
                "[azure]  You are now logged in to Azure")

    def test_logout(self, log, runner, isolated, monkeypatch, auger_authenticated):
        result = runner.invoke(cmdl, ['auth', 'logout', '-p', 'azure'])
        assert result.exit_code == 0
        assert log.records[-1].message == "[azure]  You are logged out of Azure."

    def test_whoami_anonymous(self, log, runner, monkeypatch):
        result = runner.invoke(cmdl, ['auth', 'whoami', '-p', 'azure'])
        #assert result.exit_code != 0
        assert (log.records[-1].message ==
                "[azure]  Please login to Azure...")

    def test_whoami_authenticated(self, log, runner, monkeypatch, auger_authenticated):
        azure_dict = {
          "subscription_id": "test_subscription_id",
          "directory_tenant_id": "test_directory_tenant_id",
          "application_client_id": "test_application_client_id",
          "client_secret": "test_client_secret"
        }
        AZURE_JSON = json.dumps(azure_dict)
        monkeypatch.setenv("AZURE_CREDENTIALS", AZURE_JSON)
        result = runner.invoke(cmdl, ['auth', 'whoami', '-p', 'azure'])
        assert result.exit_code == 0
        assert (log.records[-1].message ==
                "[azure]  subscription_id: test_subscription_id directory_tenant_id:test_directory_tenant_id application_client_id:test_application_client_id client_secret:test_client_secret")

    def test_logout_not_logged(self, log, runner, isolated, monkeypatch):
        monkeypatch.setenv("AZURE_CREDENTIALS", '{}')
        result = runner.invoke(cmdl, ['auth', 'logout', '-p', 'azure'])
        assert (log.records[-1].message == '[azure]  You are logged out of Azure.')
        #assert result.exit_code != 0



from .base_interpreter import BaseInterpreter
from .lexer import AstError, Lexer
from .parser import Parser, TopNode

class VarNamesFetcher(BaseInterpreter):
    def __init__(self, expression):
        self.expression = expression

    def fetch(self):
        parser = Parser(Lexer(self.expression))
        root = parser.parse()

        self.var_names = set()
        self.evaluate(root)
        return list(self.var_names)

    def evaluate_no_op_node(self, node):
        pass

    def evaluate_const_node(self, node):
        pass

    def evaluate_var_node(self, node):
        self.var_names.add(node.name)

    def evaluate_var_def_node(self, node):
        pass

    def evaluate_binary_op_node(self, node):
        self.evaluate(node.left)
        self.evaluate(node.right)

    def evaluate_unary_op_node(self, node):
        self.evaluate(node.node)

    def evaluate_func_node(self, node):
        list(map(lambda node: self.evaluate(node), node.arg_nodes))

    def evaluate_tuple_node(self, node):
        list(map(lambda node: self.evaluate(node), node.item_nodes))

    def evaluate_top_node(self, node):
        list(map(lambda n: self.evaluate(n), node.child_nodes()))

    def evaluate_with_node(self, node):
        list(map(lambda n: self.evaluate(n), node.with_item_nodes))

    def evaluate_with_item_node(self, node):
        nodes = [node.source_node]

        if node.alias_node:
            nodes.append(node.alias_node)

        list(map(lambda n: self.evaluate(n), nodes))



from .lexer import AstError, Token

class ParserError(AstError):
    pass

class BaseNode:
    def __init__(self, token):
        self.token = token
        self.require_aggregation = False

    def position(self):
        return self.token.position

class NoOpNode(BaseNode):
    def __str__(self):
        return ""

class ConstNode(BaseNode):
    def __init__(self, token):
        super().__init__(token)
        self.value = token.value

    def __str__(self):
        if isinstance(self.value, str):
            return '"' + str(self.value) + '"'
        else:
            return str(self.value)

class VarNode(BaseNode):
    def __init__(self, token):
        super().__init__(token)
        self.name = token.value

    def __str__(self):
        return str(self.name)

class VarDefNode(VarNode):
    pass

class BinaryOpNode(BaseNode):
    def __init__(self, left, token, right):
        super().__init__(token)
        self.left = left
        self.op = token.type
        self.right = right

    def __str__(self):
        return f"({self.left} {self.op} {self.right})"

class UnaryOpNode(BaseNode):
    def __init__(self, token, node):
        super().__init__(token)
        self.op = token.type
        self.node = node

    def __str__(self):
        return f"({self.op} {self.node})"

class FuncNode(BaseNode):
    def __init__(self, token, arg_nodes):
        super().__init__(token)
        self.func_name = token.value
        self.arg_nodes = arg_nodes

    def __str__(self):
        return str(self.func_name) + "(" + ", ".join(map(str, self.arg_nodes)) + ")"

class TupleNode(BaseNode):
    def __init__(self, token, items):
        super().__init__(token)
        self.item_nodes = items

    def __str__(self):
        return "(" + ", ".join(map(str, self.item_nodes))  + ")"

class TopNode(BaseNode):
    def __init__(self, token):
        super().__init__(token)
        self.kind = token.value # (top|bottom|all)
        self.limit_node = None
        self.order_node = None
        self.group_node = None
        self.having_node = None
        self.where_node = None
        self.with_node = None
        self.nested_node = None
        self.require_aggregation = True

    def child_nodes(self):
        nodes = [
            self.nested_node, # validate nested nodes first
            self.with_node, # then with which can define new vars
            self.limit_node, # then rest
            self.order_node,
            self.group_node,
            self.having_node,
            self.where_node,
        ]

        return [node for node in nodes if node is not None]

    def __str__(self):
        res = [self.kind]

        if self.kind != Token.ALL:
            res.append(str(self.limit_node))
            res.append(Token.BY)
            res.append(str(self.order_node))

        if self.with_node:
            res.append(Token.WITH)
            res.append(str(self.with_node))

        if self.group_node:
            res.append(Token.PER)
            res.append(str(self.group_node))

            if self.having_node:
                res.append(Token.HAVING)
                res.append(str(self.having_node))

        if self.where_node:
            res.append(Token.WHERE)
            res.append(str(self.where_node))

        if self.nested_node:
            res.append("from (" + str(self.nested_node) + ")")

        return " ".join(res)

class WithNode(BaseNode):
    def __init__(self, token, with_item_nodes):
        super().__init__(token)
        self.with_item_nodes = with_item_nodes

    def __str__(self):
        return ", ".join(map(str, self.with_item_nodes))

class WithItemNode(BaseNode):
    def __init__(self, token, source_node, alias_node=None):
        super().__init__(token)
        self.token = token.value
        self.source_node = source_node
        self.alias_node = alias_node

        self.require_aggregation = True
        self.source_node.require_aggregation = True

    def alias(self):
        if self.alias_node:
            return str(self.alias_node)
        else:
            return str(self.source_node)

    def __str__(self):
        res = [str(self.source_node)]

        if self.alias_node:
            res.append(Token.AS)
            res.append(str(self.alias_node))

        return " ".join(res)


class Parser:
    CONSTANTS = set([Token.CONST, Token.STR_CONST, Token.INT_CONST, Token.FLOAT_CONST])

    COMPARE_OP_BITWISE_OR_PAIR_OPS = set(
        [
            Token.EQ,
            Token.EQ2,
            Token.NE,
            Token.LTE,
            Token.LT,
            Token.GTE,
            Token.GT,
            Token.NOT,
            Token.IN
        ]
    )

    def __init__(self, lexer):
        self.lexer = lexer
        self.current_token = self.lexer.next_token()

    def error(self, msg=None):
        if self.current_token.type == Token.EOF:
            raise ParserError(f"unexpected end of expression at position {self.current_token.position}")
        else:
            msg = msg or "invalid token"
            value = self.current_token.value
            pos = self.current_token.position
            raise ParserError(f"{msg} '{value}' at position {pos}")

    def eat(self, token_type):
        # compare the current token type with the passed token
        # type and if they match then "eat" the current token
        # and assign the next token to the self.current_token,
        # otherwise raise an exception.
        if self.current_token.type == token_type:
            self.prev_token = self.current_token
            self.current_token = self.lexer.next_token()
        else:
            self.error()

    def parse(self):
        root = self.disjunction()

        if self.current_token.type != Token.EOF:
            self.error()

        return root

    def disjunction(self):
        left = self.conjunction()

        while self.current_token.type == Token.OR:
            self.eat(Token.OR)
            left = BinaryOpNode(left, self.prev_token, self.conjunction())

        return left

    def conjunction(self):
        left = self.inversion()

        while self.current_token.type == Token.AND:
            self.eat(Token.AND)
            left = BinaryOpNode(left, self.prev_token, self.inversion())

        return left

    def inversion(self):
        if self.current_token.type == Token.NOT:
            self.eat(Token.NOT)
            return UnaryOpNode(self.prev_token, self.inversion())
        else:
            return self.comparison()

    def comparison(self):
        left = self.bitwise_or()

        if self.current_token.type in Parser.COMPARE_OP_BITWISE_OR_PAIR_OPS:
            left = self.compare_op_bitwise_or_pair(left)

        return left

    def compare_op_bitwise_or_pair(self, left):
        if self.current_token.type in (Token.EQ, Token.EQ2):
            return self.eq_bitwise_or(left)
        elif self.current_token.type == Token.NE:
            return self.noteq_bitwise_or(left)
        elif self.current_token.type == Token.LTE:
            return self.lte_bitwise_or(left)
        elif self.current_token.type == Token.LT:
            return self.lt_bitwise_or(left)
        elif self.current_token.type == Token.GTE:
            return self.gte_bitwise_or(left)
        elif self.current_token.type == Token.GT:
            return self.gt_bitwise_or(left)
        elif self.current_token.type == Token.NOT:
            return self.notin_bitwise_or(left)
        if self.current_token.type == Token.IN:
            return self.in_bitwise_or(left)
        else:
            raise self.error("unknown op")

    def eq_bitwise_or(self, left):
        self.eat(Token.EQ2)
        return BinaryOpNode(left, self.prev_token, self.bitwise_or())

    def noteq_bitwise_or(self, left):
        self.eat(Token.NE)
        return BinaryOpNode(left, self.prev_token, self.bitwise_or())

    def lte_bitwise_or(self, left):
        self.eat(Token.LTE)
        return BinaryOpNode(left, self.prev_token, self.bitwise_or())

    def lt_bitwise_or(self, left):
        self.eat(Token.LT)
        return BinaryOpNode(left, self.prev_token, self.bitwise_or())

    def gte_bitwise_or(self, left):
        self.eat(Token.GTE)
        return BinaryOpNode(left, self.prev_token, self.bitwise_or())

    def gt_bitwise_or(self, left):
        self.eat(Token.GT)
        return BinaryOpNode(left, self.prev_token, self.bitwise_or())

    def notin_bitwise_or(self, left):
        self.eat(Token.NOT)
        return UnaryOpNode(self.prev_token, self.in_bitwise_or(left))

    def in_bitwise_or(self, left):
        self.eat(Token.IN)
        return UnaryOpNode(self.prev_token, self.bitwise_or())

    def bitwise_or(self):
        left = self.bitwise_xor()

        while self.current_token.type == Token.BIT_OR:
            self.eat(Token.BIT_OR)
            left = BinaryOpNode(left, self.prev_token, self.bitwise_xor())

        return left

    def bitwise_xor(self):
        left = self.bitwise_and()

        while self.current_token.type == Token.BIT_XOR:
            self.eat(Token.BIT_XOR)
            left = BinaryOpNode(left, self.prev_token, self.bitwise_and())

        return left

    def bitwise_and(self):
        left = self.shift_expr()

        while self.current_token.type == Token.BIT_AND:
            self.eat(Token.BIT_AND)
            left = BinaryOpNode(left, self.prev_token, self.shift_expr())

        return left

    def shift_expr(self):
        left = self.sum()

        while self.current_token.type in (Token.BIT_LSHIFT, Token.BIT_RSHIFT):
            self.eat(self.current_token.type)
            left = BinaryOpNode(left, self.prev_token, self.sum())

        return left

    def sum(self):
        left = self.term()

        while self.current_token.type in (Token.PLUS, Token.MINUS):
            self.eat(self.current_token.type)
            left = BinaryOpNode(left, self.prev_token, self.term())

        return left

    def term(self):
        left = self.factor()

        while self.current_token.type in (Token.MUL, Token.DIV, Token.INT_DIV, Token.MODULO):
            self.eat(self.current_token.type)
            left = BinaryOpNode(left, self.prev_token, self.factor())

        return left

    def factor(self):
        if self.current_token.type in (Token.PLUS, Token.MINUS, Token.BIT_NOT):
            self.eat(self.current_token.type)
            return UnaryOpNode(self.prev_token, self.factor())
        else:
            return self.power()

    def power(self):
        base = self.primary()

        if self.current_token.type == Token.POWER:
            self.eat(Token.POWER)
            base = BinaryOpNode(base, self.prev_token, self.factor())

        return base

    def expression(self):
        return self.disjunction()

    def primary(self):
        if self.current_token.type == Token.LPAREN:
            token = self.current_token
            self.eat(Token.LPAREN)
            expr = self.expression()
            if self.current_token.type == Token.COMMA:
                return self.tuple(token, expr)
            else:
                self.eat(Token.RPAREN)
                return expr
        elif self.current_token.type in (Token.TOP, Token.BOTTOM, Token.ALL):
            return self.top_expression()
        else:
            return self.atom()

    def tuple(self, start_token, first_item):
        items = [first_item]
        while self.current_token.type == Token.COMMA:
            self.eat(Token.COMMA)
            items.append(self.expression())

        self.eat(Token.RPAREN)
        return TupleNode(start_token, items)


    def atom(self):
        if self.current_token.type == Token.ID:
            self.eat(Token.ID)
            if self.current_token.type == Token.LPAREN:
                return self.func_call_statement()
            else:
                return VarNode(self.prev_token)
        elif self.current_token.type in Parser.CONSTANTS:
            return self.const_node()
        else:
            self.error("unknown atom")

    def const_node(self, token_type=None):
        token = self.current_token
        self.eat(token_type or self.current_token.type)
        return ConstNode(token)

    def var_def_node(self):
        token = self.current_token
        self.eat(self.current_token.type)
        return VarDefNode(token)

    def func_call_statement(self):
        name_token = self.prev_token
        self.eat(Token.LPAREN)

        arg_nodes = []

        if self.current_token.type != Token.RPAREN:
            arg_nodes.append(self.expression())

            while self.current_token.type == Token.COMMA:
                self.eat(Token.COMMA)
                arg_nodes.append(self.expression())

        self.eat(Token.RPAREN)

        return FuncNode(name_token, arg_nodes)

    def top_expression(self):
        node = TopNode(self.current_token)

        if node.kind == Token.ALL:
            self.eat(Token.ALL)
        else:
            if node.kind == Token.TOP:
                self.eat(Token.TOP)
            else:
                self.eat(Token.BOTTOM)

            node.limit_node = self.const_node(Token.INT_CONST)

            self.eat(Token.BY)
            node.order_node = self.shift_expr()

        if self.current_token.type == Token.WITH:
            self.eat(Token.WITH)
            node.with_node = self.with_node()

        if self.current_token.type == Token.PER:
            self.eat(Token.PER)
            node.group_node = self.expression()
            if self.current_token.type == Token.HAVING:
                self.eat(Token.HAVING)
                node.having_node = self.expression()

        if self.current_token.type == Token.WHERE:
            self.eat(Token.WHERE)
            node.where_node = self.expression()

        if self.current_token.type == Token.FROM:
            self.eat(Token.FROM)
            self.eat(Token.LPAREN)
            node.nested_node = self.top_expression()
            self.eat(Token.RPAREN)

        return node

    def with_node(self):
        with_items = []
        with_token = self.prev_token

        while True:
            item_token = self.current_token
            source_node = self.shift_expr()
            alias_node = None

            if self.current_token.type == Token.AS:
                self.eat(Token.AS)
                alias_node = self.var_def_node()

            with_items.append(WithItemNode(item_token, source_node, alias_node))

            if self.current_token.type == Token.COMMA:
                self.eat(Token.COMMA)
            else:
                break

        return WithNode(with_token, with_items)



import click
from a2ml.api.a2ml import A2ML
from a2ml.api.utils.context import pass_context

@click.command('import', short_help='Import data for training.')
@click.option('--source', '-s', type=click.STRING, required=False,
    help='Source file to import.If skipped, then import source from config.yml.')
@click.option('--name', '-n', type=click.STRING, required=False,
    help='Name file to import.')
@click.option('--description', '-d', type=click.STRING, required=False,
    help='Description of dataset.')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def cmdl(ctx, source, name, description, provider):
    """Import data for training."""
    ctx.setup_logger(format='')
    A2ML(ctx, provider).import_data(source, name, description=description)



from a2ml.api.base_a2ml import BaseA2ML
from a2ml.api.utils.show_result import show_result
#from a2ml.api.utils import convert_source
from a2ml.api.utils.context import Context


class A2ML(BaseA2ML):
    """Facade to A2ML providers."""

    def __init__(self, ctx, provider = None):
        """Initializes A2ML PREDIT instance.

        Args:
            ctx (object): An instance of the a2ml Context.
            provider (str): The automl provider(s) you wish to run. For example 'auger,azure,google'. The default is None - use provider set in config.

        Returns:
            A2ML object

        Examples:
            .. code-block:: python

                ctx = Context()
                a2ml = A2ML(ctx, 'auger, azure')
        """
        super(A2ML, self).__init__(ctx, None)
        self.runner = self.build_runner(ctx, provider)
        self.local_runner = lambda: self.build_runner(ctx, provider, force_local=True)

    @show_result
    def import_data(self, source=None, name=None, description=None):
        """Imports data defined in context. Uploading the same file name will result in versions being appended to the file name.

        Note:
            Your context points to a config file where ``source`` is defined.

            .. code-block:: yaml

                # Local file name, remote url to the data source file or postgres url
                source: './dataset.csv'

            .. code-block:: yaml

                # Postgres url parameters: dbname, tablename, offset(OPTIONAL), limit(OPTIONAL)
                source: jdbc:postgresql://user:pwd@ec2-54-204-21-226.compute-1.amazonaws.com:5432/dbname?tablename=table1&offset=0&limit=100

        Args:
            source (str, optional): Local file name, remote url to the data source file, Pandas DataFrame or postgres url
            name (str, optional): Name of dataset, if none then file name used. If source is DataFrame then name should be specified.
            description (str, optionsl): Description of dataset

        Returns:
            Results for each provider. ::

                {
                    'auger': {'result': True, 'data': {'created': 'dataset.csv'}}, \n
                    'azure': {'result': True, 'data': {'created': 'dataset.csv'}}
                }

            Errors. ::

                {
                    'auger': {'result': False, 'data': 'Please specify data source file...'}, \n
                    'azure': {'result': False, 'data': 'Please specify data source file...'}
                }


        Examples:
            .. code-block:: python

                ctx = Context()
                a2ml = A2ML(ctx, 'auger, azure')
                a2ml.import_data()
        """
        return self.runner.execute('import_data', source=source, name=name, description=description)

    @show_result
    def preprocess_data(self, data, preprocessors, locally=False):
        """Preprocess data

        Args:
            data (str|pandas.DataFrame): Input data for preprocess. Can be path to file(local or s3) or Pandas Dataframe
            preprocessors (array of dicts): List of preprocessors with parameters ::

                [
                    {'text': {'text_cols': []}}
                ]

        Preprocessors:
            text
                * text_cols(array): List of text columns to process
                * text_metrics ['mean_length', 'unique_count', 'separation_score'] : Calculate metrics for text fields and after vectorize(separation_score) 
                * tokenize (dict): Default - {'max_text_len': 30000, 'tokenizers': ['sent'], 'remove_chars': '○•'}
                * vectorize ('en_use_lg'|'hashing'|'en_use_md'|'en_use_cmlm_md'|'en_use_cmlm_lg'): See see https://github.com/MartinoMensio/spacy-universal-sentence-encoder
                * dim_reduction(dict): Generate features based on vectors. See https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html ::

                    {
                        'alg_name': 'PCA'|'t-SNE', 
                        'args': {'n_components': 2} #Number of components to keep.
                    }

                * output_prefix (str): Prefix for generated columns. Format name: {prefix}_{colname}_{num}

                * calc_distance ['none', 'cosine', 'cityblock', 'euclidean', 'haversine', 'l1', 'l2', 'manhattan', 'nan_euclidean'] | 'cosine' : See https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.distance_metrics.html#sklearn.metrics.pairwise.distance_metrics
                * compare_pairs (array of dicts): When calc_distance is not none. ::

                    [
                        {'compare_cols': [{'dataset_idx': 0, 'cols': ['col1']}, {'dataset_idx': 1, 'cols': ['col2']}], 
                            'output_name':'cosine_col1_col2', 'params': {}
                        },
                        {'compare_cols': [{'dataset_idx': 0, 'cols': ['col3']}, {'dataset_idx': 1, 'cols': ['col4']}], 
                            'output_name':'cosine_col3_col4', 'params': {}
                        },                        
                    ]

                * datasets: List of datasets to process, may be empty, so all fields takes from main dataset ::

                    [
                        {'path': 'path', 'keys': ['main_key', 'local_key'], 'text_metrics': ['separation_score', 'mean_length', 'unique_count']}, 
                        {'path': 'path1', 'keys': ['main_key1', 'local_key1']}
                    ]

        Returns:
                {
                    'result': True,
                    'data': 'data in input format'
                }

        """
        return self.get_runner(locally).execute_one_provider('preprocess_data', data, preprocessors, locally)

    @show_result
    def train(self):
        """Starts training session based on context state.

        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'data': {
                            'experiment_name': 'dataset.csv-4-experiment',
                            'session_id': '9ccfe04eca67757a'
                         }
                    },
                    'azure': {
                        'result': True,
                        'data': {
                            'experiment_name': 'dataset.csv-4-experiment',
                            'session_id': '9ccfe04eca67757a'
                         }
                    }
                }

            Errors. ::

                {
                    'auger': {'result': False, 'data': 'Please set target to build model.'}, \n
                    'azure': {'result': False, 'data': 'Please set target to build model.'}
                }

        Examples:

            .. code-block:: python

                ctx = Context()
                a2ml = A2ML(ctx, 'auger, azure')
                a2ml.train()

        """
        return self.runner.execute('train')

    @show_result
    def evaluate(self, run_id = None):
        """Evaluate the results of training.

        Args:
            run_id (str, optional): The run id for a training session. A unique run id is created for every train. Default is last experiment train.
        
        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'data': {
                            'run_id': '9ccfe04eca67757a',
                            'leaderboard': [
                                {'model id': 'A017AC8EAD094FD', 'rmse': '0.0000', 'algorithm': 'LGBMRegressor'},
                                {'model id': '4602AFCEEEAE413', 'rmse': '0.0000', 'algorithm': 'ExtraTreesRegressor'}
                            ],
                            'trials_count': 10,
                            'status': 'started',
                            'provider_status': 'provider specific'
                        }
                    },
                    'azure': {
                        'result': True,
                        'data': {
                            'run_id': '9ccfe04eca67757a',
                            'leaderboard': [
                                {'model id': 'A017AC8EAD094FD', 'rmse': '0.0000', 'algorithm': 'LGBMRegressor'},
                                {'model id': '4602AFCEEEAE413', 'rmse': '0.0000', 'algorithm': 'ExtraTreesRegressor'}
                            ],
                            'trials_count': 10,
                            'status': 'started',
                            'provider_status': 'provider specific'
                        }
                    }
                }

            **Status**

                * **preprocess** - search is preprocessing data for traing
                * **started** - search is in progress
                * **completed** - search is completed
                * **interrupted** - search was interrupted
                * **error** - search was finished with error

        Examples:
            .. code-block:: python

                ctx = Context()
                a2ml = A2ML(ctx, 'auger, azure')
                while True:
                    res = a2ml.evaluate()
                    if status['auger']['status'] not in ['preprocess','started']:
                        break
        """
        return self.runner.execute('evaluate', run_id = run_id)

    @show_result
    def deploy(self, model_id, locally=False, review=False, provider=None,
            name=None, algorithm=None, score=None, data_path=None, metadata=None):
        """Deploy a model locally or to specified provider(s).

        Note:
            See evaluate function to get model_id \n
            This method support only one provider

        Args:
            model_id (str): The model id from any experiment you will deploy. Ignored for 'external' provider
            locally(bool): Deploys the model locally if True, on the Provider Cloud if False. The default is False.
            review(bool): Should model support review based on actual data. The default is True.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider defined by model_id or set in costructor.
            name (str): Friendly name for the model. Used as name for Review Endpoint
            algorithm (str): Monitored model(external provider) algorithm name.
            score (float): Monitored model(external provider) score.
            data_path (str): Data path to fit model when deploy. Return new deployed model-id
            metadata (dict): Additional parameter for the model. Used for accurcay report(report parameter)

        Returns:
            ::

                {
                    'result': True,
                    'data': {'model_id': 'A017AC8EAD094FD'}
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                a2ml = A2ML(ctx, 'auger, azure')
                a2ml.deploy(model_id='A017AC8EAD094FD', name='FirstExperiment')

            .. code-block:: python

                ctx = Context()
                a2ml = A2ML(ctx, 'external')
                result = a2ml.deploy(model_id=None, name="My external model.", algorithm='RandomForest', score=0.75)
                model_id = result['data']['model_id']

        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('deploy',
            model_id, locally, review, name, algorithm, score, data_path, metadata)

    @show_result
    def predict(self, model_id, filename=None, data=None, columns=None, predicted_at=None,
            threshold=None, score=False, score_true_data=None,
            output=None, no_features_in_result = None, locally=False, provider=None):
        """Predict results with new data against deployed model. Predictions are stored next to the file with data to be predicted on. The file name will be appended with suffix _predicted.

        Note:
            Use deployed model_id \n
            This method support only one provider

        Args:
            model_id(str): The deployed model id you want to use.
            filename(str): The file with data to request predictions for.
            data: array of records [[target, actual]] or Pandas DataFrame (target, actual) or dict created with Pandas DataFrame to_dict('list') method
            columns(list): list of column names if data is array of records
            predicted_at: Predict data date. Use for review of historical data.
            threshold(float): For classification models only. This will return class probabilities with response.
            score(bool): Calculate scores for predicted results.
            score_true_data(str, pandas.DataFrame, dict): Data with true values to calculate scores. If missed, target from filename used for true values.
            output(str): Output csv file path.
            no_features_in_result(bool) : Do not return feature columns in prediction result. False by default
            locally(bool, str): Predicts using a local model with auger.ai.predict if True, on the Provider Cloud if False. If set to "docker", then docker image used to run the model
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider set in costructor or config.

        Returns:
            if filename is not None. ::

                {
                    'result': True,
                    'data': {'predicted': 'dataset_predicted.csv'}
                }

            if filename is None and data is not None and columns is None. ::

                {
                    'result': True,
                    'data': {'predicted': [{col1: value1, col2: value2, target: predicted_value1}, {col1: value3, col2: value4, target: predicted_value2}]}
                }

            if filename is None and data is not None and columns is not None. ::

                {
                    'result': True,
                    'data': {'predicted': {'columns': ['col1', 'col2', target], 'data': [['value1', 'value2', 1], ['value3', 'value4', 0]]}}
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                rv = A2ML(ctx).predict(model_id, '../irises.csv')
                # if rv[provider].result is True
                # predictions are stored in rv[provider]['data']['predicted']

            .. code-block:: python

                ctx = Context()
                data = [{'col1': 'value1', 'col2': 'value2'}, {'col1': 'value3', 'col2': 'value4'}]
                rv = A2ML(ctx).predict(model_id, data=data)
                # if rv[provider].result is True
                # predictions are returned as rv[provider]['data']['predicted']

            .. code-block:: python

                ctx = Context()
                data = [['value1', 'value2'], ['value3', 'value4']]
                columns = ['col1', 'col2']
                rv = A2ML(ctx).predict(model_id, data=data)
                # if rv[provider].result is True
                # predictions are returned as rv[provider]['data']['predicted']

            .. code-block:: python
                
                # Predict locally without config files. Model will automatically downloaded if not exists.
                # To use local predict install a2ml[predict]
                ctx = Context()
                ctx.config.set('name', 'project name')
                ctx.credentials = "Json string from a2ml ui settings"

                rv = A2ML(ctx).predict(model_id, '../irises.csv',
                    no_features_in_result = True, locally=True)
                # if rv[provider].result is True
                # predictions are stored in rv[provider]['data']['predicted']

        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('predict', filename, model_id, 
            threshold, locally, data, columns, predicted_at, output, no_features_in_result,
            score, score_true_data )

    @show_result
    def actuals(self, model_id, filename=None, data=None, columns=None, actuals_at=None, 
            actual_date_column=None, experiment_params=None, locally=False, provider=None):
        """Submits actual results(ground truths) for predictions of a deployed model. This is used to review and monitor active models.

        Note:
            It is assumed you have predictions against this model first.

            .. list-table:: actuals.csv
                :widths: 50 50 50
                :header-rows: 1

                * - predicted( or target): predicted value. If missed - predict called automatically
                  - actual
                  - baseline_target: predicted value for baseline model (OPTIONAL)
                * - Iris-setosa
                  - Iris-setosa
                  - Iris-setosa
                * - Iris-virginica
                  - Iris-virginica
                  - Iris-virginica

            It may also contain train features to predict(if target missed), retrain model while Review and for distribution chart

            This method support only one provider

        Args:
            model_id(str): The deployed model id you want to use.
            filename(str): The file with data to request predictions for.
            data: array of records [[target, actual]] or Pandas DataFrame (target, actual) or dict created with Pandas DataFrame to_dict('list') method
            columns(list): list of column names if data is array of records
            actuals_at: Actuals date. Use for review of historical data.
            actual_date_column(str): name of column in data which contains actual date
            experiment_params(dict): parameters to calculate experiment metrics ::

                start_date(date): experiment actuals start date
                end_date(date):  experiment actuals end date
                date_col(str): column name with date

            locally(bool): Process actuals locally.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider set in costructor or config.

        Returns:
            ::

                {
                    'result': True,
                    'data': True
                }

            Errors. ::

                {
                    'result': False,
                    'data': 'Actual Prediction IDs not found in model predictions.'
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                A2ML(ctx).actuals('D881079E1ED14FB', filename=<path_to_file>/actuals.csv)

            .. code-block:: python

                ctx = Context()
                actual_records = [['predicted_value_1', 'actual_value_1'], ['predicted_value_2', 'actual_value_2']]
                columns = [target, 'actual']

                A2ML(ctx).actuals('D881079E1ED14FB', data=actual_records,columns=columns)

            .. code-block:: python

                ctx = Context()
                actual_records = [['predicted_value_1', 'actual_value_1'], ['predicted_value_2', 'actual_value_2']]
                columns = [target, 'actual']

                A2ML(ctx, "external").actuals('external_model_id', data=actual_records,columns=columns)

        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('actuals', 
            model_id, filename, data, columns, actuals_at, actual_date_column, experiment_params, locally)


    @show_result
    def delete_actuals(self, model_id, with_predictions=False, begin_date=None, end_date=None, locally=False, provider=None):
        """Delete files with actuals and predcitions locally or from specified provider(s).

        Args:
            model_id (str): Model ID to delete actuals and predictions.
            with_predictions(bool):
            begin_date: Date to begin delete operations
            end_date: Date to end delete operations
            locally(bool): Delete files from local model if True, on the Provider Cloud if False. The default is False.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider defined by model_id or set in costructor.

        Returns:
            ::

                {
                    'result': True,
                    'data': None
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                A2MLModel(ctx).delete_actuals(model_id='D881079E1ED14FB')
        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('delete_actuals', model_id, with_predictions, begin_date, end_date, locally)


    @show_result
    def review(self, model_id, locally=False, provider=None):
        """Review information about deployed model.

        Args:
            model_id(str): The deployed model id you want to use.
            locally(bool): Process review locally.

        Returns:
            status(str): May be : started, error, completed, retrain
            error(str): Description of error if status='error'
            accuracy(float): Average accuracy of model(based on used metric) for review sensitivity period(see config.yml)

            ::

                {
                    'result': True,
                    'data': {'status': 'completed', 'error': '', 'accuracy': 0.76}
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                result = A2ML(ctx).review(model_id='D881079E1ED14FB')
        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('review', model_id)




import os
import pytest
from a2ml.api.a2ml import A2ML
from a2ml.api.google.a2ml import GoogleA2ML
from a2ml.api.utils.context import Context

class TestGoogleA2ML():

    @pytest.mark.skip(reason='make it work first')
    def test_init(self):
        fulldir=os.getcwd()+"/tests/test_google"
        os.chdir(fulldir)
        # load config(s) from test app
        ctx = Context()

        a2ml = A2ML(ctx)
        assert len(a2ml.runner.providers)==3
        assert isinstance(a2ml.runner.providers[0], GoogleA2ML)
    
    @pytest.mark.skip(reason='make it work first')
    def test_import_data(self): 

        # load config(s) from test app
        self.ctx = Context()

        a2ml = GoogleA2ML(self.ctx)
        result = a2ml.import_data()
        assert(len(result.dataset_id)>0)
    
    @pytest.mark.skip(reason='make it work first')
    def test_train(self):
        ctx = Context()
        a2ml = GoogleA2ML(ctx)
        result = a2ml.train()
        print("Train result: {}",format(result))
 


